package dr.evomodel.antigenic.phyloclustering.misc.obsolete;

public class ClusterComparison {
 public int minNumMismatchPerfectSeparation;
 public int sampleSize;
 public double SS;
 public double scaledMinNumMismatchPerfectSeparation;
}
