package dr.evomodel.arg.operators;

/**
 * @author Andrew Rambaut
 * @version $Id$
 */
public class ARGOperatorFailedException extends Exception {
    public ARGOperatorFailedException(String message) {
        super(message);
    }
}
