/*
 * ComparableDouble.java
 *
 * Copyright (c) 2002-2015 Alexei Drummond, Andrew Rambaut and Marc Suchard
 *
 * This file is part of BEAST.
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership and licensing.
 *
 * BEAST is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 *  BEAST is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with BEAST; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */

package dr.util;

/** 
 * This class is unfortunate but necessary to conform to JDK 1.1
 *
 * @version $Id: ComparableDouble.java,v 1.3 2005/05/24 20:26:01 rambaut Exp $
 *
 * @author Alexei Drummond
 */
public class ComparableDouble implements Comparable {
	
	private final double value;

	public ComparableDouble(double d) {
		value = d;
	}

	public int compareTo(Object o) {

		ComparableDouble cd = (ComparableDouble)o;

		if (value < cd.value) {
			return -1;
		} else if (value > cd.value) {
			return 1;
		} else return 0;
	}

	public boolean equals(Object o) {

		ComparableDouble cd = (ComparableDouble)o;
		return cd.value == value;
	}

	public double doubleValue() {
		return value;
	}

	public String toString() {
		return value + "";
	}
}
