/*
 * Polygon2DFactory.java
 *
 * Copyright (c) 2002-2017 Alexei Drummond, Andrew Rambaut and Marc Suchard
 *
 * This file is part of BEAST.
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership and licensing.
 *
 * BEAST is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 *  BEAST is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with BEAST; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */

package dr.geo;

import org.jdom.Attribute;
import org.jdom.Element;

import java.awt.geom.Point2D;
import java.util.LinkedList;

/**
 * @author Guy Baele
 */
public class Polygon2DFactory {

    public static AbstractPolygon2D createPolygon2D(Element element) {
        //Attribute attrib = element.getAttribute("fillValue");
        Attribute samplingAttrib = element.getAttribute("samplingProbability");
        if (samplingAttrib != null) {
            double fillValue = Double.parseDouble(samplingAttrib.getValue());
            return new Polygon2DSampling(element, fillValue);
        }
        else {
            return new Polygon2D(element);
        }
    }

    public static AbstractPolygon2D createPolygon2D(LinkedList<Point2D> clippedPolygon, boolean closed) {
        return new Polygon2D(clippedPolygon, closed);
    }

}
