/*
 * CompoundFunction.java
 *
 * Copyright (c) 2002-2015 Alexei Drummond, Andrew Rambaut and Marc Suchard
 *
 * This file is part of BEAST.
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership and licensing.
 *
 * BEAST is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 *  BEAST is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with BEAST; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */

package dr.math;


/**
 * interface for a function of one variable
 *
 * @author Korbinian Strimmer
 */
public class CompoundFunction implements UnivariateFunction {
	
	public CompoundFunction(UnivariateFunction[] functions, double Z) {
		this.functions = functions;
		this.Z = Z;
	}
		
	public double evaluate(double x) {
		double value = Z;
		for (int i =0; i < functions.length; i++) {
			value *= functions[i].evaluate(x);
		}
		return value;
	}
	
	public double getLowerBound() { return functions[0].getLowerBound(); }
	public double getUpperBound() { return functions[0].getUpperBound(); }
		
	UnivariateFunction[] functions = null;	
	double Z = 1.0;
}
